rm(list=ls())
# this is the initial random seed
set.seed(1234)
# this program serves to estimate genomic selection
# SNP effects and then use these estimates to predict the breeding value of 
# another set of idividuals
# we see the predictive ability by the correlation between estimated breeding value and 
# observed phenotype
# by doing multiple cross-validation
# mice data set, body length
# prepared for a course in UBA, Buenos Aires, February 2009
# Updated for another course in UAB, Barcelona, September 2010
# UGA, Athens, GA, may 2012
# Andres Legarra, INRA 

solve_densem_pcg_X <- function(X,rhs,vara,vare){
   # solve [X'X +(I vara)-1] sol = rhs by preconditioned conjugate gradient for densem A
   # it includes an overall mean, the rest are SNPs
   # all effects are random but the mean

   n=length(rhs)
   m=rep(0,n)
   De=rep(1/vara,n)
   De[1]=0
   for (i in 1:n){
     m[i]=1/( crossprod(X[,i])/vare + De[i])
   }
   sol=rep(0,n)
   p=rep(0,n)
   z=rep(0,n)
   w=rep(0,n)
   r=rhs
  
   for (k in 1:1000){
     z=m*r
     tau=crossprod(z,r)
      if (k == 1){
         beta=0
	       p=z
       }else{
         beta=as.numeric(tau/oldtau)
         p=z+beta*p
      }
      w=crossprod(X,X%*%p)/vare + De*p
      alpha=as.numeric(tau/crossprod(p,w))
      sol=sol+alpha*p
      if ((k%%100) != 0){
           r=r-alpha*w
         }else{
           r=rhs-crossprod(X,X%*%sol)/vare - De*sol
      }
      conv=crossprod(r,r)/crossprod(rhs,rhs)
      #print(c('round ',k,'   convergence=',conv))
      if (conv < 1e-14) break
      oldtau=tau
   }   
   #print(c(k,' iterations,   convergence criterion=',conv))
   sol
}


GSRU <- function(y,X,vare,vara,ahat=rep(0,dim(X)[2])){
  # this function solves a blup for genomic selection
  # given design matrix X, data y, residual variance vare, SNP variance vara
  # solve by Gauss Seidel
  neq=dim(X)[2]
  e=y-X%*%ahat
  #ahat=rep(0,neq)
  xpx=rep(0,neq)
  epsit=rep(0,1)
  # this is GSRU, legarra & Misztal, Journal of Dairy Science 2008
  #set up diag(X'X)
  for (i in 1:neq) {
    xpx[i]=crossprod(X[,i],X[,i])
  }
  mu=0
  # do until convergence
  for (iter in 1:1000) {
    #Gauss Seidel
    eps=0
    for (i in 1:neq){
      lhs=xpx[i]/vare
      if(i>1) lhs=lhs+1/vara
      rhs=crossprod(X[,i],e)/vare + xpx[i]/vare*ahat[i]
      val=rhs/lhs
      eps=eps+((val - ahat[i])**2)
      e = e - X[,i]*(val - ahat[i])
      ahat[i]=val
    }
    if(iter%%10==0) print(c(iter,eps))
    epsit[iter]=eps
    if(eps<1E-10) break
  }
  print(c(iter,eps,date()))
  ahat
}  

# ---------------------------------------------------
# 
# ---------------------------------------------------

set=1  #10946
   #2  #5473
   #3  #1095

if(set==1)
    # this is the data set with all SNPs
    a=matrix( scan('bl10946.txt'),nrow=1829,byrow=TRUE)
    vara=9.09E-06 # this is for the whole 10946 snps
if(set==2)
    # this one has 1 every 2 SNPs
    a=matrix( scan('bl5473.txt'),nrow=1829,byrow=TRUE)
    vara=4.5E-04 # this is for 5473 SNPs
if(set==3)
    # this one has 1 every 10 SNPs
    a=matrix( scan('bl1095.txt'),nrow=1829,byrow=TRUE)
    vara=1.02E-04 # this is for 1095 snps

# extract design matrix for SNPs (exclude id, phenotype and sex code)
X=a[,-(1:3)]
#center design matrix
for (i in 1:ncol(X)){
  X[,i]=X[,i]-mean(X[,i])
}

#extract data
y=a[,2]
# add design matrix for general mean
X=cbind(1,X)
vare=0.200
print(c("vara",vara))
system.time({ # let's check if this is long
  #a = GSRU(y,X,vare,vara) # very long !!
  a = solve_densem_pcg_X(X,crossprod(X,y)/vare,vara,vare)
})
yhat=X%*%a
print(cor(y,yhat))

# ---------------------------------------------------
# END OF FIRST PART
# ---------------------------------------------------

# now do the same by splitting the data into two
i=0
test=seq(200,1600,by=200)
ntest=10
pp=matrix(0,test*ntest)
size=matrix(0,test*ntest)
pos=0
for (ind_in_training in test){ 
  i=i+1
  for (j in 1:ntest){
    pos=pos+1
    # here we select some animals to be the "training" data set
    training=sample(1:1829,ind_in_training,replace=TRUE)
    # tho others will be used for validation
    valid=(1:1829)[-training]
    # we estimate SNP effects with individuals in "training" 
    # excluding the ones in validation
    #a = GSRU(y[training],X[training,],vare,vara,a)
    a=solve_densem_pcg_X(X[training,],crossprod(X[training,],y[training])/vare,vara,vare)
    # using estimates for a, here we predict phenotypes for everybody
    yhat=X%*%a
    # how much is the accuracy for animals in validation?
    print(c(ind_in_training,j,cor(y[valid],yhat[valid])))
    # how much is the accuracy for animals in training?
    #print(c(cor(y[training],yhat[training]),i,j))
    pp[pos]=cor(y[valid],yhat[valid])
    size[pos]=ind_in_training
  }
}

boxplot(pp ~size)

# ---------------------------------------------------
# END OF SECOND PART
# ---------------------------------------------------
 
   

